{
  StudioGM(TM) Sofwtare (C) 1997-2009
  Copyright (C) Grusha Maksim A.
  http://www.studiogm.us
  studiogm@studiogm.us

  Version 1.3
  Componsnts: TSGMEditListBox

  Created: 2007-08-01
  Updated: 2009-05-12

  2009-05-12
  version 1.3
    * Fixed methods:
      + CMMouseWheel;
      + DoEndEditItem;
  2009-05-08
  version 1.2
    + Added property AutoOptions default is [aoEdit, aoEditOnDblClick].
    * Fixed events:
      + OnDrawItem;
      + OnMeasureItem;
      + OnBeginEditItem
      + OnEndEditItem;
    * Fixed methods:
      + OnDrawItem;
      + OnMeasureItem;
      + DoBeginEditItem;
      + DoEndEditItem;
  2009-05-04
  version 1.1
    * Added inplace editor class.
    * Added methods:
      + BeginUpdate;
      + EndUpdate;
    * Added properties:
      + Colors
      + UpdateCount
    * Fixed methods:
      + CreateWnd;
      + DrawItem;
      + DeleteItem;

  2007-08-01
  version 1.0
    * First release.
    * Added methods:
      + DeleteItem;
      + DoBeginEditItem;
      + DoEndEditItem;
      + InsertItemBefore;
      + InsertItemAfter;
    * Added properties:
      + ReadOnly;
      + ShowFocusedRect;
    * Added events:
      + OnBeginEditItem;
      + OnEndEditItem;
      + OnChangeItem;
}
unit usgmEditListBox;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls,
  Forms, Math;
  
const
  SPalettePage = 'StudioGM(TM)';

type
  TSGMCustomEditListBox = class;

  TOnChangeItem = procedure (Sender: TObject; LastIndex, NewIndex: Integer;
    var CanChange: Boolean) of object;

  TOnBeginEditItem = procedure (Sender: TObject; Index: Integer;
    var CanEdit: Boolean; var SelStart, SelLength: Integer) of object;

  TOnEndEditItem = procedure( Sender: TObject; Index: Integer;
    var NewText: String; const OriginalText: String;
    var EndEdit: Boolean ) of object;


  TSGMEditListBoxInplaceEditor = class(TCustomEdit)
  private
    FListBox: TSGMCustomEditListBox;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
  public
    constructor Create(AOwner: TComponent); override;
    property ListBox: TSGMCustomEditListBox read FListBox;
  end;

  TSGMAutoOption = (aoDelete, aoEdit, aoEditOnDblClick, aoInsert);
  TSGMAutoOptions = set of TSGMAutoOption;

  TSGMCustomEditListBoxColors = class(TPersistent)
  private
    FColors: array[0..6] of TColor;
    FListBox: TSGMCustomEditListBox;
    function GetColor(const Index: Integer): TColor;
    procedure SetColor(const Index: Integer; const Value: TColor);
  protected
    procedure Changed;
  public
    constructor Create(AListBox: TSGMCustomEditListBox);
    property ListBox: TSGMCustomEditListBox read FListBox;
  published
    property Normal:        TColor index 0 read GetColor write SetColor default clWhite;
    property Selected:      TColor index 1 read GetColor write SetColor default $00C6943A;
    property Innactive:     TColor index 2 read GetColor write SetColor default $00DBDBDB;
    property TextNormal:    TColor index 3 read GetColor write SetColor default clBlack;
    property TextSelected:  TColor index 4 read GetColor write SetColor default clwhite;
    property TextInnactive: TColor index 5 read GetColor write SetColor default $0057504A;
  end;

  TSGMCustomEditListBox = class( TCustomListBox )
  private
    FOnChangeItem: TOnChangeItem;
    FOnBeginEditItem: TOnBeginEditItem;
    FOnEndEditItem: TOnEndEditItem;
    FShowFocusedRect: Boolean;
    FReadOnly: Boolean;
    FColors: TSGMCustomEditListBoxColors;
    FAutoOptions: TSGMAutoOptions;
    procedure SetShowFocusedRect( const Value: Boolean );
    procedure CMMouseWheel(var Message: TCMMouseWheel); message CM_MOUSEWHEEL;
    procedure WMEraseBkgnd( var Message: TWmEraseBkgnd ); message WM_ERASEBKGND;
    procedure WMLButtonDown( var Message: TWMLButtonDown ); message WM_LBUTTONDOWN;
    procedure WMPaint( var Message: TWMPaint ); message WM_PAINT;
    procedure WMHScroll( var Message: TWMHScroll ); message WM_HSCROLL;
    procedure WMVScroll( var Message: TWMVScroll ); message WM_VSCROLL;
  protected
    //
    FEditIndex: Integer;
    FLastIndex: Integer;
    FEdit: TSGMEditListBoxInplaceEditor;
    FEditMode: Boolean;
    FEditRect: TRect;
    FUpdateCount: Integer;
    //
    procedure Click; override;
    procedure CreateWnd; override;
    procedure DblClick; override;
    procedure DestroyWnd; override;
    procedure DoChangeItem;
    procedure DrawItem( Index: Integer; Rect: TRect; State: TOwnerDrawState ); override;
    procedure KeyDown( var Key: Word; Shift: TShiftState ); override;
    procedure MeasureItem( Index: Integer; var Height: Integer ); override;
    procedure Resize; override;
    //
    procedure UpdateBackground;
    //
    property AutoOptions: TSGMAutoOptions
      read FAutoOptions
      write FAutoOptions default [ aoEdit, aoEditOnDblClick ];
    property Colors: TSGMCustomEditListBoxColors
      read FColors;
    property ReadOnly: Boolean
      read FReadOnly
      write FReadOnly default False;
    property ShowFocusedRect: Boolean
      read FShowFocusedRect
      write SetShowFocusedRect default False;
    property OnBeginEditItem: TOnBeginEditItem
      read FOnBeginEditItem
      write FOnBeginEditItem;
    property OnEndEditItem: TOnEndEditItem
      read FOnEndEditItem
      write FOnEndEditItem;
    property OnChangeItem: TOnChangeItem
      read FOnChangeItem
      write FOnChangeItem;
  public
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate( NeedInvalidate: Boolean = True );
    procedure DeleteItem( Index: Integer );
    procedure DoBeginEditItem;
    procedure DoEndEditItem( Canceled: Boolean = False );
    procedure InsertItemBefore( const Text: String = '' );
    procedure InsertItemAfter( const Text: String = '' );
    property EditMode: Boolean read FEditMode;
    property EditIndex: Integer read FEditIndex; 
    property UpdateCount: Integer read FUpdateCount;
  end;

  TSGMEditListBox = class(TSGMCustomEditListBox)
  published
    property AutoOptions;
    property AutoComplete;
    property Align;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelKind default bkNone;
    property BevelOuter;
    property BiDiMode;
    property BorderStyle;
    property Colors;
    property Columns;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property ExtendedSelect;
    property Font;
    property ImeMode;
    property ImeName;
    property IntegralHeight;
    property ItemHeight;
    property Items;
    property MultiSelect;
    property ParentBiDiMode;
    //property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ScrollWidth;
    property ShowFocusedRect;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property TabWidth;
    property Visible;
    property OnBeginEditItem;
    property OnChangeItem;
    property OnClick;
    property OnContextPopup;
    property OnData;
    property OnDataFind;
    property OnDataObject;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnEndDock;
    property OnEndDrag;
    property OnEndEditItem;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;  

implementation

{$R *.RES}

procedure Register;
begin
  RegisterComponents(SPalettePage, [TSGMEditListBox]);
end;

{ TSGMEditListBoxInplaceEditor }

constructor TSGMEditListBoxInplaceEditor.Create(AOwner: TComponent);
begin
  inherited Create( AOwner );
  AutoSelect := False;
  if Assigned(AOwner) and AOwner.InheritsFrom(TSGMCustomEditListBox) then
    FListBox := AOwner as TSGMCustomEditListBox
  else
    FListBox := nil;
end;

procedure TSGMEditListBoxInplaceEditor.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if Assigned(FListBox) then
    case Key of
      VK_UP:
      begin
        FListBox.DoEndEditItem;
        if FListBox.ItemIndex > 0 then
          FListBox.ItemIndex := FListBox.ItemIndex - 1;
      end;
      VK_DOWN:
      begin
        FListBox.DoEndEditItem;
        if FListBox.ItemIndex < FListBox.Count - 1 then
          FListBox.ItemIndex := FListBox.ItemIndex + 1;
      end;
    end;
  inherited KeyDown(Key, Shift);
end;

procedure TSGMEditListBoxInplaceEditor.KeyPress(var Key: Char);
begin
  if Assigned(FListBox) then
    case Key of
      #27:
      begin
        Text := FListBox.Items[ FListBox.ItemIndex ];
        FListBox.DoEndEditItem( True );
      end;
      #13: FListBox.DoEndEditItem;
    end;
  inherited KeyPress( Key );
end;

procedure TSGMEditListBoxInplaceEditor.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result or DLGC_WANTALLKEYS;
end;

{ TSGMCustomEditListBoxColors }

procedure TSGMCustomEditListBoxColors.Changed;
begin
  if Assigned(FListBox) then InvalidateRect(FListBox.Handle, nil, False);
end;

constructor TSGMCustomEditListBoxColors.Create(AListBox: TSGMCustomEditListBox);
begin
  FListBox := AListBox;
  FColors [ 0 ] := clWhite;   // Normal
  FColors [ 1 ] := $00C6943A; // Selected
  FColors [ 2 ] := $00DBDBDB; // Innactive
  FColors [ 3 ] := clBlack;   // TextNormal
  FColors [ 4 ] := clWhite;   // TextSelected
  FColors [ 5 ] := $0057504A; // TextInnactive
end;

function TSGMCustomEditListBoxColors.GetColor(const Index: Integer): TColor;
begin
  Result := FColors[ Index ];
end;

procedure TSGMCustomEditListBoxColors.SetColor(const Index: Integer; const Value: TColor);
begin
  if FColors[ Index ] = Value then Exit;
  FColors[ Index ] := Value;
  Changed;
end;

{ TSGMCustomEditListBox }

procedure TSGMCustomEditListBox.Click;
begin
  inherited Click;
  if FLastIndex = ItemIndex then Exit;
  DoChangeItem;
end;

constructor TSGMCustomEditListBox.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FAutoOptions := [ aoEdit, aoEditOnDblClick ];
  FColors := TSGMCustomEditListBoxColors.Create( Self );
  FUpdateCount := 0;
  FLastIndex := -1;
  FEditMode := False;
  FEditIndex := -1;
  FDoubleBuffered := True;
  FReadOnly := False;
  FShowFocusedRect := False;
  Style := lbOwnerDrawVariable;
end;

procedure TSGMCustomEditListBox.CreateWnd;
begin
  inherited CreateWnd;
  FEdit := TSGMEditListBoxInplaceEditor.Create( Self );
  FEdit.Parent := Self;
  FEdit.Visible := False;
  FEdit.BorderStyle := bsNone;
  FEdit.SetBounds(-1, -1, 0, 0);
  FEditRect := Rect(0, 0, 0, 0);
end;

procedure TSGMCustomEditListBox.DblClick;
begin
  inherited DblClick;
  if not FEditMode and (ItemIndex > -1) and (aoEditOnDblClick in FAutoOptions)
  then
    DoBeginEditItem;
end;

procedure TSGMCustomEditListBox.DestroyWnd;
begin
  FreeAndNil(FEdit);
  inherited DestroyWnd;
end;

procedure TSGMCustomEditListBox.DoBeginEditItem;
var
  SelStart, SelLength: Integer;
begin
  if (ItemIndex < 0) or not Assigned( FEdit ) or FReadOnly  then Exit;
  FEditMode := True;
  SelStart := 0;
  SelLength := Length( Items[ ItemIndex ] );
  if Assigned(FOnBeginEditItem) then
    FOnBeginEditItem(
      Self,
      ItemIndex,
      FEditMode,
      SelStart,
      SelLength
    );
  if not FEditMode then Exit;
  BeginUpdate;
  try
    FEditIndex := ItemIndex;
    FEdit.BoundsRect := FEditRect;
    FEdit.Text := Items[ ItemIndex ];
    FEdit.Visible := True;
    FEdit.SetFocus;
    FEdit.SelStart := SelStart;
    FEdit.SelLength := SelLength;
  finally
    EndUpdate;
  end;
end;

procedure TSGMCustomEditListBox.DoChangeItem;
var
  CanChange: Boolean;
begin
  CanChange := True;

  if FEditMode then DoEndEditItem;

  if FEditMode and ((FLastIndex = -1) or ((FLastIndex > -1)
    and (FLastIndex < Items.Count))) then
  begin
    ItemIndex := FLastIndex;
    Exit;
  end;

  if Assigned(FOnChangeItem) then
    FOnChangeItem(Self, FLastIndex, ItemIndex, CanChange);

  if CanChange then
    FLastIndex := ItemIndex
  else if (FLastIndex = -1) or ((FLastIndex > -1)
    and (FLastIndex < Items.Count)) then
      ItemIndex := FLastIndex;
end;

procedure TSGMCustomEditListBox.DoEndEditItem( Canceled: Boolean = False );
var
  NewText, OriginalText: String;
  EndEdit: Boolean;
begin
  if not FEditMode or not Assigned( FEdit ) then Exit;

  FEditMode := False;
  EndEdit := True;

  OriginalText := Items[ ItemIndex ];
  NewText := FEdit.Text;

  Canceled := Canceled or (AnsiCompareStr(NewText, OriginalText) = 0);

  if Assigned( FOnEndEditItem ) and not Canceled then
    FOnEndEditItem( Self, ItemIndex, NewText, OriginalText,  EndEdit );

  if EndEdit or Canceled then
  begin
    if csLButtonDown in ControlState then
      ControlState := ControlState - [ csLButtonDown ];
  end;

  if EndEdit or Canceled then
  begin
    FEditMode := False;
    FEdit.Visible := False;
    FEditIndex := -1;
    if not Canceled then Items[ ItemIndex ] := NewText;
    SetFocus;
    Invalidate;
  end;

end;

procedure TSGMCustomEditListBox.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  S: String;
  R: TRect;
  OldBkMode: Integer;
begin
  TControlCanvas(Canvas).UpdateTextFlags;

  if ItemIndex = Index then
  begin
    FEditRect := Rect;
    InflateRect(FEditRect, -2, -2);
  end;

  if Assigned(OnDrawItem) then
  begin
    OnDrawItem(Self, Index, Rect, State);
    Exit;
  end;
  
  OldBkMode := SetBkMode(Canvas.Handle, TRANSPARENT);

  if (Index <> FEditIndex)  then
  begin
    if FEditMode then
    begin
      Canvas.Brush.Color := Colors.Innactive;
      Canvas.Font.Color := Colors.TextInnactive;
    end
    else
    begin
      if odSelected in State then
      begin
        Canvas.Brush.Color := Colors.Selected;
        Canvas.Font.Color := Colors.TextSelected;
      end
      else
      begin
        Canvas.Brush.Color := Colors.Normal;
        Canvas.Font.Color := Colors.TextNormal;
      end;
    end;
    Canvas.FillRect(Rect);
    S := Items[ Index ];
    if Length(S) > 0 then
    begin
      Canvas.TextOut(
        5,
        Rect.Top + ((Rect.Bottom-Rect.Top) shr 1) - (Canvas.TextHeight(S) shr 1),
        S
      );
    end;
  end
  else
  begin
    R := Rect;
    Inc( R.Top );
    Inc( R.Left );
    Canvas.Brush.Style := bsClear;
    Canvas.Pen.Color := clBlack;
    Canvas.Pen.Width := 2;
    Canvas.Rectangle( R );
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Width := 1;
  end;
  //
  if (not FShowFocusedRect and (odFocused in State)) then
    Canvas.DrawFocusRect(Rect);
  // Restore last BkMode
  SetBkMode(Canvas.Handle, OldBkMode);
end;

procedure TSGMCustomEditListBox.InsertItemBefore(const Text: String);
begin
  if FReadOnly then Exit;
  if FEditMode then DoEndEditItem;
  Items.Insert(ItemIndex, Text);
  ItemIndex := Max(0, Pred(ItemIndex));
  DoBeginEditItem;
end;

procedure TSGMCustomEditListBox.InsertItemAfter(const Text: String);
begin
  if FReadOnly then Exit;
  if FEditMode then DoEndEditItem;
  Items.Insert(Min(Count, Succ(ItemIndex)), Text);
  ItemIndex := Max(0, Succ(ItemIndex));
  DoBeginEditItem;
end;

procedure TSGMCustomEditListBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_DELETE, VK_SUBTRACT: if aoDelete in FAutoOptions then DeleteItem(ItemIndex);
    VK_INSERT, VK_ADD: if aoInsert in FAutoOptions then InsertItemAfter;
    VK_F2: if aoEdit in FAutoOptions then DoBeginEditItem;
   else
    inherited KeyDown(Key, Shift);
  end;
end;

procedure TSGMCustomEditListBox.MeasureItem(Index: Integer; var Height: Integer);
begin
  Canvas.Font.Assign( Font );
  Height := Math.Max(
    ItemHeight,
    Canvas.TextHeight('Iq') + 4
  );
  if Assigned(OnMeasureItem) then
    inherited MeasureItem(Index, Height);
end;

procedure TSGMCustomEditListBox.SetShowFocusedRect(const Value: Boolean);
begin
  if FShowFocusedRect = Value then Exit;
  FShowFocusedRect := Value;
  Invalidate;
end;

procedure TSGMCustomEditListBox.WMLButtonDown(var Message: TWMLButtonDown);
begin
  if FEditMode and not PtInRect(FEditRect, SmallPointToPoint(Message.Pos)) then
    DoEndEditItem;
  FLastIndex := ItemIndex;
  inherited;
end;

procedure TSGMCustomEditListBox.DeleteItem(Index: Integer);
var
  NewIndex: Integer;
begin
  if FReadOnly then Exit;
  if FEditMode then DoEndEditItem;
  if (Index > -1) and (Index < Count) then
  begin
    if (ItemIndex = Index) and (Count > 1) then
    begin
      NewIndex := Max(0, Pred(Index));
      if (NewIndex = 0) and (Index = 0) then Inc(NewIndex);
      ItemIndex := NewIndex;
    end;
    Items.Delete(Index);
  end;
end;

procedure TSGMCustomEditListBox.WMPaint(var Message: TWMPaint);
begin
  if FUpdateCount > 0 then Exit;
  if Message.DC <> 0 then UpdateBackground;
  inherited;
end;

procedure TSGMCustomEditListBox.WMEraseBkgnd(var Message: TWmEraseBkgnd);
begin
  inherited;
end;

destructor TSGMCustomEditListBox.Destroy;
begin
  FreeAndNil(FColors);
  inherited Destroy;
end;

procedure TSGMCustomEditListBox.WMHScroll(var Message: TWMHScroll);
begin
  if FEditMode then DoEndEditItem;
  inherited;
end;

procedure TSGMCustomEditListBox.WMVScroll(var Message: TWMVScroll);
begin
  if FEditMode then DoEndEditItem;
  inherited;
end;

procedure TSGMCustomEditListBox.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TSGMCustomEditListBox.EndUpdate(NeedInvalidate: Boolean);
begin
  if FUpdateCount = 0 then Exit;
  Dec(FUpdateCount);
  if FUpdateCount > 0 then Exit;
  if NeedInvalidate then Invalidate;
end;

procedure TSGMCustomEditListBox.UpdateBackground;
begin
  if FEditMode then
    Canvas.Brush.Color := Colors.Innactive
  else
    Canvas.Brush.Color := Colors.Normal;
  Canvas.FillRect( ClientRect );
end;

procedure TSGMCustomEditListBox.Resize;
begin
  if FEditMode then DoEndEditItem;
  inherited Resize;
end;

procedure TSGMCustomEditListBox.CMMouseWheel(var Message: TCMMouseWheel);
begin
  if FEditMode then DoEndEditItem;
  inherited;
  Invalidate;
end;

end.
